@@ -15,13 +15,14 @@ namespace TransactionalUpdate {
 
 class Plugins {
 public:
-    Plugins(TransactionalUpdate::Transaction* transaction);
+    Plugins(TransactionalUpdate::Transaction* transaction, bool ignore_error);
     virtual ~Plugins();
     void run(std::string stage, std::string args);
     void run(std::string stage, char* argv[]);
 protected:
     TransactionalUpdate::Transaction* transaction;
     std::vector<std::filesystem::path> plugins;
+    bool ignore_error;
 };
 
 } // namespace TransactionalUpdate
