@@ -42,8 +42,10 @@ class Transaction::impl {
 public:
     void addSupplements();
     void snapMount();
+    void closeSnapshot(bool aborted=false);
     int runCommand(char* argv[], bool inChroot, std::string* buffer);
     static int inotifyAdd(const char *pathname, const struct stat *sbuf, int type, struct FTW *ftwb);
+    static int selinux_logging_callback(int type, const char *fmt, ...);
     int inotifyRead();
     std::unique_ptr<SnapshotManager> snapshotMgr;
     std::unique_ptr<Snapshot> snapshot;
@@ -51,6 +53,7 @@ public:
     std::vector<std::unique_ptr<Mount>> dirsToMount;
     Supplements supplements;
     pid_t pidCmd;
+    bool keepIfError = false;
     bool discardIfNoChange = false;
 };
 
@@ -79,8 +82,12 @@ Transaction::~Transaction() {
     try {
         if (isInitialized() && !getSnapshot().empty() && fs::exists(getRoot())) {
             tulog.info("Discarding snapshot ", pImpl->snapshot->getUid(), ".");
-            pImpl->snapshot->abort();
-            TransactionalUpdate::Plugins plugins{nullptr};
+            if (pImpl->keepIfError) {
+                pImpl->closeSnapshot(true);
+            } else {
+                pImpl->snapshot->abort();
+            }
+            TransactionalUpdate::Plugins plugins{nullptr, pImpl->keepIfError};
             plugins.run("abort-post", pImpl->snapshot->getUid());
         }
     }  catch (const std::exception &e) {
@@ -104,6 +111,28 @@ fs::path Transaction::getBindDir() {
     return pImpl->bindDir;
 }
 
+int Transaction::impl::selinux_logging_callback(int type, const char *fmt, ...) {
+    va_list ap, ap_copy;
+    va_start(ap, fmt);
+    va_copy(ap_copy, ap);
+    size_t size = vsnprintf(nullptr, 0, fmt, ap);
+    va_end(ap);
+    std::vector<char> se_buffer(size + 1);
+    vsnprintf(se_buffer.data(), size, fmt, ap_copy);
+
+    switch (type) {
+        case SELINUX_ERROR:
+            tulog.error("SELinux: ", se_buffer.data());
+            break;
+        default:
+            tulog.info("SELinux: ", se_buffer.data());
+            break;
+    }
+    va_end(ap_copy);
+
+    return 0;
+}
+
 void Transaction::impl::snapMount() {
     if (unshare(CLONE_NEWNS) < 0) {
         throw std::runtime_error{"Creating new mount namespace failed: " + std::string(strerror(errno))};
@@ -134,8 +163,10 @@ void Transaction::impl::snapMount() {
             // up in the root file system, but will always be shadowed by the real /var mount. Due to that they
             // also won't be relabelled at any time. During updates this may cause problems if packages try to
             // access those leftover directories with wrong permissions, so they have to be relabelled manually...
-            BindMount selinuxVar("/var/lib/selinux", 0, true);
-            selinuxVar.mount(bindDir);
+            if (fs::is_directory("/var/lib/selinux")) {
+                BindMount selinuxVar("/var/lib/selinux", 0, true);
+                selinuxVar.mount(bindDir);
+            }
             BindMount selinuxEtc("/etc/selinux", 0, true);
             selinuxEtc.mount(bindDir);
 
@@ -148,6 +179,10 @@ void Transaction::impl::snapMount() {
                     tulog.error("Chrooting to " + bindDir.native() + " for SELinux relabelling failed: " + std::string(strerror(errno)));
                     _exit(errno);
                 }
+
+                union selinux_callback se_callback;
+                se_callback.func_log = selinux_logging_callback;
+                selinux_set_callback(SELINUX_CB_LOG, se_callback);
                 unsigned int restoreconOptions = SELINUX_RESTORECON_RECURSE | SELINUX_RESTORECON_IGNORE_DIGEST;
                 if (tulog.level >= TULogLevel::Info)
                     restoreconOptions |= SELINUX_RESTORECON_VERBOSE;
@@ -177,7 +212,8 @@ void Transaction::impl::snapMount() {
     mntRun->setType("tmpfs");
     mntRun->setSource("tmpfs");
     dirsToMount.push_back(std::move(mntRun));
-    dirsToMount.push_back(std::make_unique<BindMount>("/run/systemd/journal"));
+    if (fs::exists("/run/systemd/journal"))
+        dirsToMount.push_back(std::make_unique<BindMount>("/run/systemd/journal"));
     std::unique_ptr<Mount> mntVarTmp{new Mount{"/var/tmp"}};
     mntVarTmp->setType("tmpfs");
     mntVarTmp->setSource("tmpfs");
@@ -256,7 +292,7 @@ int Transaction::impl::inotifyAdd(const char *pathname, const struct stat *sbuf,
 }
 
 void Transaction::init(std::string base, std::optional<std::string> description) {
-    TransactionalUpdate::Plugins plugins{nullptr};
+    TransactionalUpdate::Plugins plugins{nullptr, pImpl->keepIfError};
     plugins.run("init-pre", nullptr);
 
     if (base == "active")
@@ -275,7 +311,7 @@ void Transaction::init(std::string base, std::optional<std::string> description)
         std::unique_ptr<Snapshot> prevSnap = pImpl->snapshotMgr->open(base);
         std::unique_ptr<Mount> oldEtc{new Mount{"/etc"}};
         oldEtc->setTabSource(prevSnap->getRoot() / "etc" / "fstab");
-        if (oldEtc->getFilesystem() == "overlayfs") {
+        if (oldEtc->isMount() && oldEtc->getFilesystem() == "overlayfs") {
             tulog.info("Can not merge back changes in /etc into old overlayfs system - ignoring 'discardIfNoChange'.");
         } else {
             // Flag file to indicate this snapshot was initialized with discard flag
@@ -285,12 +321,12 @@ void Transaction::init(std::string base, std::optional<std::string> description)
         }
     }
 
-    TransactionalUpdate::Plugins plugins_with_transaction{this};
+    TransactionalUpdate::Plugins plugins_with_transaction{this, pImpl->keepIfError};
     plugins_with_transaction.run("init-post", nullptr);
 }
 
 void Transaction::resume(std::string id) {
-    TransactionalUpdate::Plugins plugins{nullptr};
+    TransactionalUpdate::Plugins plugins{nullptr, pImpl->keepIfError};
     plugins.run("resume-pre", id);
 
     pImpl->snapshot = pImpl->snapshotMgr->open(id);
@@ -304,10 +340,14 @@ void Transaction::resume(std::string id) {
         pImpl->discardIfNoChange = true;
     }
 
-    TransactionalUpdate::Plugins plugins_with_transaction{this};
+    TransactionalUpdate::Plugins plugins_with_transaction{this, pImpl->keepIfError};
     plugins_with_transaction.run("resume-post", nullptr);
 }
 
+void Transaction::setKeepIfError(bool keep) {
+    pImpl->keepIfError = keep;
+}
+
 void Transaction::setDiscardIfUnchanged(bool discard) {
     pImpl->discardIfNoChange = discard;
 }
@@ -341,7 +381,10 @@ int Transaction::impl::runCommand(char* argv[], bool inChroot, std::string* outp
 
         // Recursively register all directories of the root file system
         inotifyExcludes = MountList::getList(snapshot->getRoot());
-        inotifyExcludes.push_back(snapshot->getRoot() / "etc");
+        std::unique_ptr<Mount> mntEtc{new Mount{"/etc"}};
+        if (mntEtc->isMount()) {
+            inotifyExcludes.push_back(snapshot->getRoot() / "etc");
+        }
         nftw(snapshot->getRoot().c_str(), inotifyAdd, 20, FTW_MOUNT | FTW_PHYS);
     }
 
@@ -458,7 +501,7 @@ int Transaction::impl::runCommand(char* argv[], bool inChroot, std::string* outp
 }
 
 int Transaction::execute(char* argv[], std::string* output) {
-    TransactionalUpdate::Plugins plugins{this};
+    TransactionalUpdate::Plugins plugins{this, pImpl->keepIfError};
     plugins.run("execute-pre", argv);
     int status = this->pImpl->runCommand(argv, true, output);
     plugins.run("execute-post", argv);
@@ -477,7 +520,7 @@ int Transaction::callExt(char* argv[], std::string* output) {
         argv[i] = strdup(s.c_str());
     }
 
-    TransactionalUpdate::Plugins plugins{this};
+    TransactionalUpdate::Plugins plugins{this, pImpl->keepIfError};
     plugins.run("callExt-pre", argv);
     int status = this->pImpl->runCommand(argv, false, output);
     plugins.run("callExt-post", argv);
@@ -492,14 +535,11 @@ void Transaction::sendSignal(int signal) {
     }
 }
 
-void Transaction::finalize() {
-    TransactionalUpdate::Plugins plugins{this};
-    plugins.run("finalize-pre", nullptr);
-
+void Transaction::impl::closeSnapshot(bool aborted) {
     sync();
-    if (pImpl->discardIfNoChange &&
-            ((inotifyFd != 0 && pImpl->inotifyRead() == 0) ||
-            (inotifyFd == 0 && fs::exists(getRoot() / "discardIfNoChange")))) {
+    if (discardIfNoChange &&
+            ((inotifyFd != 0 && inotifyRead() == 0) ||
+            (inotifyFd == 0 && fs::exists(snapshot->getRoot() / "discardIfNoChange")))) {
         tulog.info("No changes to the root file system - discarding snapshot.");
 
         // Even if the snapshot itself does not contain any changes, /etc may do so. If the new snapshot is a
@@ -511,51 +551,62 @@ void Transaction::finalize() {
             std::filesystem::path targetRoot = "/";
             std::string base;
 
-            std::ifstream input(getRoot() / "discardIfNoChange");
+            std::ifstream input(snapshot->getRoot() / "discardIfNoChange");
             input >> base;
             input.close();
 
             std::unique_ptr<Mount> previousEtc{new Mount("/etc", 0, true)};
-            if (pImpl->snapshotMgr->getCurrent() == base) {
+            if (snapshotMgr->getCurrent() == base) {
                 tulog.info("Merging changes in /etc into the running system.");
             } else {
                 tulog.info("Merging changes in /etc into the previous snapshot.");
-                targetRoot = pImpl->snapshotMgr->open(base)->getRoot();
+                targetRoot = snapshotMgr->open(base)->getRoot();
             }
-            Util::exec("rsync --archive --inplace --xattrs --acls --exclude 'fstab' --exclude 'etc.syncpoint' --delete --quiet '" + this->pImpl->bindDir.native() + "/etc/' " + targetRoot.native() + "/etc");
+            Util::exec("rsync --archive --inplace --xattrs --acls --exclude 'fstab' --exclude 'etc.syncpoint' --delete --quiet '" + bindDir.native() + "/etc/' " + targetRoot.native() + "/etc");
         }
 
-        TransactionalUpdate::Plugins plugins_without_transaction{nullptr};
-        plugins_without_transaction.run("finalize-post", pImpl->snapshot->getUid() + " " + "discarded");
+        TransactionalUpdate::Plugins plugins_without_transaction{nullptr, keepIfError};
+        plugins_without_transaction.run("finalize-post", snapshot->getUid() + " " + "discarded");
         return;
     }
-    if (fs::exists(getRoot() / "discardIfNoChange")) {
-        fs::remove(getRoot() / "discardIfNoChange");
+    if (fs::exists(snapshot->getRoot() / "discardIfNoChange")) {
+        fs::remove(snapshot->getRoot() / "discardIfNoChange");
     }
 
     // Update /usr timestamp to support system offline update mechanism
-    if (utime((pImpl->snapshot->getRoot() / "usr").c_str(), nullptr) != 0)
+    if (utime((snapshot->getRoot() / "usr").c_str(), nullptr) != 0)
         throw std::runtime_error{"Updating /usr timestamp failed: " + std::string(strerror(errno))};
 
-    pImpl->snapshot->close();
-    pImpl->supplements.cleanup();
-    pImpl->dirsToMount.clear();
+    if (! aborted) {
+        snapshot->close();
+    }
+    supplements.cleanup();
+    dirsToMount.clear();
 
-    std::unique_ptr<Snapshot> defaultSnap = pImpl->snapshotMgr->open(pImpl->snapshotMgr->getDefault());
+    std::unique_ptr<Snapshot> defaultSnap = snapshotMgr->open(snapshotMgr->getDefault());
     if (defaultSnap->isReadOnly())
-        pImpl->snapshot->setReadOnly(true);
-    pImpl->snapshot->setDefault();
-    tulog.info("New default snapshot is #" + pImpl->snapshot->getUid() + " (" + std::string(pImpl->snapshot->getRoot()) + ").");
+        snapshot->setReadOnly(true);
+    if (! aborted) {
+        snapshot->setDefault();
+        tulog.info("New default snapshot is #" + snapshot->getUid() + " (" + std::string(snapshot->getRoot()) + ").");
+    }
+}
+
+void Transaction::finalize() {
+    TransactionalUpdate::Plugins plugins{this, pImpl->keepIfError};
+    plugins.run("finalize-pre", nullptr);
+
+    this->pImpl->closeSnapshot();
 
     std::string id = pImpl->snapshot->getUid();
     pImpl->snapshot.reset();
 
-    TransactionalUpdate::Plugins plugins_without_transaction{nullptr};
+    TransactionalUpdate::Plugins plugins_without_transaction{nullptr, pImpl->keepIfError};
     plugins_without_transaction.run("finalize-post", id);
 }
 
 void Transaction::keep() {
-    TransactionalUpdate::Plugins plugins{this};
+    TransactionalUpdate::Plugins plugins{this, pImpl->keepIfError};
     plugins.run("keep-pre", nullptr);
 
     sync();
@@ -567,6 +618,6 @@ void Transaction::keep() {
     std::string id = pImpl->snapshot->getUid();
     pImpl->snapshot.reset();
 
-    TransactionalUpdate::Plugins plugins_without_transaction{nullptr};
+    TransactionalUpdate::Plugins plugins_without_transaction{nullptr, pImpl->keepIfError};
     plugins_without_transaction.run("keep-post", id);
 }
